/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.example.watcher;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Timer;
import java.util.TimerTask;

import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.app.Activity;
import android.app.Notification;
import android.app.NotificationManager;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.util.Log;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.Window;
import android.view.View.OnClickListener;
import android.view.inputmethod.EditorInfo;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;

public class MainActivity extends Activity {

    // Debugging
    private static final String TAG = "MainActivity";
    private static final boolean D = true;

    // Key names received from the BluetoothChatService Handler
    public static final String DEVICE_NAME = "device_name";
    public static final String TOAST = "toast";

    // Layout Views
    private TextView mTitle;
    private ListView mConversationView;
    private Button mConnectButton;
    private Button mDiscoverButton;
    private Button mNotificationButton;

    // Name of the connected device
    private String mConnectedDeviceName = null;
    // Array adapter for the conversation thread
    private ArrayAdapter<String> mConversationArrayAdapter;
    // Local Bluetooth adapter
    private BluetoothAdapter mBluetoothAdapter = null;
    // Member object for the Bluetooth Manager
    private BtManager mBtManager = null;
    
    // Notification broadcast receiver
    private NotificationReceiver mReceiver;
    private ArrayList<MyNotification> mNotificationList = new ArrayList<MyNotification>();
    
    
    // Auto-refresh timer
    Timer mRefreshTimer = null;
    Timer mDeleteTimer = null;

	
    /***************************************
     * Overrided, System
     ***************************************/
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
        if(D) Log.e(TAG, "+++ ON CREATE +++");

        // Set up the window layout
        requestWindowFeature(Window.FEATURE_CUSTOM_TITLE);
        setContentView(R.layout.activity_main);
        getWindow().setFeatureInt(Window.FEATURE_CUSTOM_TITLE, R.layout.custom_title);

        // Set up the custom title
        mTitle = (TextView) findViewById(R.id.title_left_text);
        mTitle.setText(R.string.app_name);
        mTitle = (TextView) findViewById(R.id.title_right_text);

        // Get local Bluetooth adapter
        mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();

        // If the adapter is null, then Bluetooth is not supported
        if (mBluetoothAdapter == null) {
            Toast.makeText(this, "Bluetooth is not available", Toast.LENGTH_LONG).show();
            finish();
            return;
        }
        
        mReceiver = new NotificationReceiver();
        IntentFilter filter = new IntentFilter();
        filter.addAction("com.example.watcher.NOTIFICATION_LISTENER");
        registerReceiver(mReceiver,filter);
	}

    @Override
    public void onStart() {
        super.onStart();
        if(D) Log.e(TAG, "++ ON START ++");

        // If BT is not on, request that it be enabled.
        // setupChat() will then be called during onActivityResult
        if (!mBluetoothAdapter.isEnabled()) {
            Intent enableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            startActivityForResult(enableIntent, Constants.REQUEST_ENABLE_BT);
        // Otherwise, setup the chat session
        } else {
            if (mBtManager == null) setupBT();
        }
    }

    @Override
    public synchronized void onResume() {
        super.onResume();
        if(D) Log.e(TAG, "+ ON RESUME +");

        // Performing this check in onResume() covers the case in which BT was
        // not enabled during onStart(), so we were paused to enable it...
        // onResume() will be called when ACTION_REQUEST_ENABLE activity returns.
        if (mBtManager != null) {
            // Only if the state is STATE_NONE, do we know that we haven't started already
            if (mBtManager.getState() == BtManager.STATE_NONE) {
              // Start the Bluetooth chat services
            	mBtManager.start();
            }
        }
        
        getAllNotifications();
    }
	
    @Override
    public synchronized void onPause() {
        super.onPause();
        if(D) Log.e(TAG, "- ON PAUSE -");
    }

    @Override
    public void onStop() {
        super.onStop();
        if(D) Log.e(TAG, "-- ON STOP --");
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
        
        // Stop the Bluetooth chat services
        if (mBtManager != null) mBtManager.stop();
        // Unregister broadcast receiver
        unregisterReceiver(mReceiver);
        // Stop the timer
        if(mRefreshTimer != null)
        	mRefreshTimer.cancel();
        if(mDeleteTimer != null)
        	mDeleteTimer.cancel();
        
        if(D) Log.e(TAG, "--- ON DESTROY ---");
    }

    // The Handler that gets information back from the BluetoothChatService
    private final Handler mHandler = new Handler() {
        @Override
        public void handleMessage(Message msg) {
            switch (msg.what) {
            case Constants.MESSAGE_STATE_CHANGE:
                if(D) Log.i(TAG, "MESSAGE_STATE_CHANGE: " + msg.arg1);
                switch (msg.arg1) {
                case BtManager.STATE_CONNECTED:
                    mTitle.setText(R.string.title_connected_to);
                    mTitle.append(mConnectedDeviceName);
                    sendNotifications();
                    
                    if(mRefreshTimer == null)
                    	mRefreshTimer = new Timer();
                    mRefreshTimer.schedule(mTimerTask, 60*60*1000, 60*60*1000);
                    break;
                case BtManager.STATE_CONNECTING:
                    mTitle.setText(R.string.title_connecting);
                    break;
                case BtManager.STATE_LISTEN:
                case BtManager.STATE_NONE:
                    mTitle.setText(R.string.title_not_connected);
                    break;
                }
                break;
            case Constants.MESSAGE_WRITE:
                // Disabled: 
            	// byte[] writeBuf = (byte[]) msg.obj;
                // construct a string from the buffer
                // String writeMessage = new String(writeBuf);
                break;
            case Constants.MESSAGE_READ:
                byte[] readBuf = (byte[]) msg.obj;
                // construct a string from the valid bytes in the buffer
                String readMessage = new String(readBuf, 0, msg.arg1);
                break;
            case Constants.MESSAGE_DEVICE_NAME:
                // save the connected device's name
                mConnectedDeviceName = msg.getData().getString(DEVICE_NAME);
                Toast.makeText(getApplicationContext(), "Connected to "
                               + mConnectedDeviceName, Toast.LENGTH_SHORT).show();
                break;
            case Constants.MESSAGE_TOAST:
                Toast.makeText(getApplicationContext(), msg.getData().getString(TOAST),
                               Toast.LENGTH_SHORT).show();
                break;
            }
        }
    };

    public void onActivityResult(int requestCode, int resultCode, Intent data) {
        if(D) Log.d(TAG, "onActivityResult " + resultCode);
        switch (requestCode) {
        case Constants.REQUEST_CONNECT_DEVICE:
            // When DeviceListActivity returns with a device to connect
            if (resultCode == Activity.RESULT_OK) {
                // Get the device MAC address
                String address = data.getExtras()
                                     .getString(DeviceListActivity.EXTRA_DEVICE_ADDRESS);
                // Get the BLuetoothDevice object
                BluetoothDevice device = mBluetoothAdapter.getRemoteDevice(address);
                // Attempt to connect to the device
                mBtManager.connect(device);
            }
            break;
        case Constants.REQUEST_ENABLE_BT:
            // When the request to enable Bluetooth returns
            if (resultCode == Activity.RESULT_OK) {
                // Bluetooth is now enabled, so set up a BT session
            	setupBT();
            } else {
                // User did not enable Bluetooth or an error occured
                Log.d(TAG, "BT not enabled");
                Toast.makeText(this, R.string.bt_not_enabled_leaving, Toast.LENGTH_SHORT).show();
                finish();
            }
        }
    }
    
    
	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.main, menu);
		return true;
	}
	
    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
        case R.id.scan:
            // Launch the DeviceListActivity to see devices and do scan
        	doScan();
            return true;
        case R.id.discoverable:
            // Ensure this device is discoverable by others
            ensureDiscoverable();
            return true;
        case R.id.noti_settings:
        	// Launch notification settings screen
        	setNotificationAccess();
        	return true;
        }
        return false;
    }

    
    /***************************************
     * Private methods 
     ***************************************/
    
    /**
     * Setup UI components and initialize BT manager
     */
    private void setupBT() {
        Log.d(TAG, "setupBT()");

        // Initialize the array adapter for the conversation thread
        mConversationArrayAdapter = new ArrayAdapter<String>(this, R.layout.message);
        mConversationView = (ListView) findViewById(R.id.in);
        mConversationView.setAdapter(mConversationArrayAdapter);

        // Initialize buttons with a listener that for click events
        mConnectButton = (Button) findViewById(R.id.button_connect);
        mConnectButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                doScan();
            }
        });
        
        mDiscoverButton = (Button) findViewById(R.id.button_discover);
        mDiscoverButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
            	ensureDiscoverable();
            }
        });
        
        mNotificationButton = (Button) findViewById(R.id.button_notification);
        mNotificationButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
            	setNotificationAccess();
            }
        });

        // Initialize the BluetoothChatService to perform bluetooth connections
        mBtManager = new BtManager(this, mHandler);
    }
    
    /**
     * Launch the DeviceListActivity to see devices and do scan
     */
    private void doScan() {
        Intent serverIntent = new Intent(this, DeviceListActivity.class);
        startActivityForResult(serverIntent, Constants.REQUEST_CONNECT_DEVICE);
    }
    
    /**
     * Launch notification settings screen
     */
    private void setNotificationAccess() {
    	Intent intent=new Intent("android.settings.ACTION_NOTIFICATION_LISTENER_SETTINGS");
    	startActivity(intent);
    }
    
    /**
     * Ensure this device is discoverable by others
     */
    private void ensureDiscoverable() {
        if(D) Log.d(TAG, "ensure discoverable");
        if (mBluetoothAdapter.getScanMode() !=
            BluetoothAdapter.SCAN_MODE_CONNECTABLE_DISCOVERABLE) {
            Intent discoverableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_DISCOVERABLE);
            discoverableIntent.putExtra(BluetoothAdapter.EXTRA_DISCOVERABLE_DURATION, 300);
            startActivity(discoverableIntent);
        }
    }

    /**
     * Sends a message.
     * @param message  A string of text to send.
     */
    private void sendMessage(String message) {
        // Check that we're actually connected before trying anything
        if (mBtManager.getState() != BtManager.STATE_CONNECTED) {
            // Toast.makeText(this, R.string.not_connected, Toast.LENGTH_SHORT).show();
            return;
        }

        // Check that there's actually something to send
        if (message.length() > 0) {
            // Get the message bytes and tell the BluetoothChatService to write
            byte[] send = message.getBytes();
            mBtManager.write(send);
        }
    }
    
    /**
     * Sends a message.
     * @param buf  Buffer to send.
     */
    private void sendMessage(byte[] buf) {
        // Check that we're actually connected before trying anything
        if (mBtManager.getState() != BtManager.STATE_CONNECTED) {
            // Toast.makeText(this, R.string.not_connected, Toast.LENGTH_SHORT).show();
            return;
        }

        // Check that there's actually something to send
        if (buf.length > 0) {
            // Get the message bytes and tell the BluetoothChatService to write
            mBtManager.write(buf);
        }
    }
    
    private void makeNotification(String title, String text, String ticker) {
        NotificationManager nManager = (NotificationManager) getSystemService(NOTIFICATION_SERVICE);
        Notification.Builder ncomp = new Notification.Builder(this);
        ncomp.setContentTitle(title);
        if(text != null)
        	ncomp.setContentText(text);
        if(ticker != null)
        	ncomp.setTicker(ticker);
        ncomp.setSmallIcon(R.drawable.ic_launcher);
        ncomp.setAutoCancel(true);
        nManager.notify((int)System.currentTimeMillis(),ncomp.build());
    }
    
    private void clearAllNotifications() {
        Intent i = new Intent("com.example.watcher.NOTIFICATION_LISTENER_SERVICE");
        i.putExtra("command","clearall");
        sendBroadcast(i);
        
        mNotificationList.clear();
        mConversationArrayAdapter.clear();
    }
    
    private void getAllNotifications() {
        Intent i = new Intent("com.example.watcher.NOTIFICATION_LISTENER_SERVICE");
        i.putExtra("command","list");
        sendBroadcast(i);
        
        mNotificationList.clear();
        mConversationArrayAdapter.clear();
    }
    
    private String[] mFilters = {"android"};
    private String[] mConverterBefore = {"gm"};
    private String[] mConverterAfter = {"Gmail"};
    
    private int sendNotifications() {
    	// send time first
    	Calendar c = Calendar.getInstance();
    	int curNoon = c.get(Calendar.AM_PM);
    	int curHour = c.get(Calendar.HOUR);
    	int curMinute = c.get(Calendar.MINUTE);
    	int curWeek = c.get(Calendar.DAY_OF_WEEK);
    	
    	byte[] t_buffer = new byte[5];
    	t_buffer[0] = '@';
    	t_buffer[1] = (byte)curNoon;
    	t_buffer[2] = (byte)curHour;
    	t_buffer[3] = (byte)curMinute;
    	t_buffer[4] = (byte)curWeek;
    	sendMessage(t_buffer);
    	
    	// send new list signal
    	sendMessage("#");
    	
    	// send package name
    	int count = 0;
    	for(MyNotification noti : mNotificationList) {
    		if( sendNotification(noti) )
    			count++;
    	}
    	
    	Toast.makeText(this, Integer.toString(count)+" notifications are sent", Toast.LENGTH_LONG).show();
    	return count;
    }
    
    private boolean sendNotification(MyNotification noti) {
		if(noti.mPackageName == null)
			return false;
		
		boolean pass = true;
		for(int i=0; i<mFilters.length; i++) {
    		if( noti.mPackageName.startsWith(mFilters[i]) )
    			pass = false;
		}

		// Filtering package name
		String trimmedName = null;
		if(pass) {
			// Trim package name
			if( noti.mPackageName.contains(".") ) {
				trimmedName = noti.mPackageName.substring(noti.mPackageName.lastIndexOf(".") + 1);
			} else {
				trimmedName = noti.mPackageName;
			}
			// Convert name
			for(int i=0; i<mConverterBefore.length; i++) {
				if(mConverterBefore[i].equals(trimmedName))
					trimmedName = mConverterAfter[i];
			}
			// Send
			sendMessage("+" + trimmedName + ";");
		}
		
		return pass;
    }
    
    /***************************************
     * Sub classes 
     ***************************************/
    
    /**
     * Broadcast receiver class to receive notification data
     */
    class NotificationReceiver extends BroadcastReceiver{

        @Override
        public void onReceive(Context context, Intent intent) {
            int cmd = intent.getIntExtra(Constants.NOTIFICATION_KEY_CMD, 0);
        	int noti_id = intent.getIntExtra(Constants.NOTIFICATION_KEY_ID, -1);
        	String packageName = intent.getStringExtra(Constants.NOTIFICATION_KEY_PACKAGE);
        	String textTicker = intent.getStringExtra(Constants.NOTIFICATION_KEY_TEXT);
            
            switch(cmd) {
            case Constants.NOTIFICATION_CMD_LIST:
            case Constants.NOTIFICATION_CMD_ADD:
            	if(packageName != null) {
            		Log.d(TAG,"** Main activity : Add noti="+noti_id+", package="+packageName);
            		
            		MyNotification myNoti = new MyNotification(noti_id, packageName, textTicker);
            		mNotificationList.add(myNoti);
            		mConversationArrayAdapter.add(myNoti.mPackageName);
            		sendNotification(myNoti);
            	}
            	break;
            	
            case Constants.NOTIFICATION_CMD_REMOVE:
        		Log.d(TAG,"** Main activity : Delete noti="+noti_id+", package="+packageName);
        		
            	for(int i = mNotificationList.size() - 1; i > -1; i--) {
            		MyNotification noti = mNotificationList.get(i);
            		if(noti.id == noti_id) {
            			mNotificationList.remove(i);
            			if(noti.mPackageName != null)
            				mConversationArrayAdapter.remove(noti.mPackageName);
            		}
            	}
            	
            	if(mDeleteTimer == null) {
            		mDeleteTimer = new Timer();
            		mDeleteTimer.schedule(mDeleteTimerTask, 5*1000);
            	}
            	break;
            }	// End of switch(cmd)
        }
    }	// End of class NotificationReceiver
    
    /**
     * Notification structure
     */
    class MyNotification {
    	int id;
    	String mPackageName = null;
    	String mText = null;
    	
    	MyNotification(int id, String packageName, String text) {
    		this.id = id;
    		this.mPackageName = packageName;
    		this.mText = text;
    	}
    }
    
    /**
     * Auto-refresh Timer
     */
    TimerTask mTimerTask = new TimerTask(){
    	public void run() {
    		mHandler.post(new Runnable() {
    			public void run() {
    				sendNotifications();
    			}
    		});
    	}
    };
    
    TimerTask mDeleteTimerTask = new TimerTask(){
    	public void run() {
    		mHandler.post(new Runnable() {
    			public void run() {
    				sendNotifications();
    				mDeleteTimer = null;
    			}
    		});
    	}
    };

}
